#!/usr/bin/perl
#
#  lav -- read an alignment file in lav or concise format, and
#         perform a variety of queries on it
#
#   Entry points:
#
#       get_align
#               Returns alignment information from an lav or concise file
#               into a new array of Segment structs. For an lav file, a Segment
#               is created for each contig block, local alignment, and gap-free
#               segment, which are interleaved in the order they appear in
#               the file (i.e., the information in a contig entry applies
#               to all of the other entries that follow it up to the next
#               contig entry, and similarly for local alignment entries).
#               The three types of entries use different subsets of the
#               available fields. Note that the endpoint values for local
#               alignments and gap-free segments are adjusted to refer to
#               the user's original full sequence files, rather than the
#               raw (possibly offset or reversed) coordinates produced by
#               blastz.  For a concise file the array is similar, except
#               that the data fields for local alignment entries are empty,
#               and endpoint values for contig entries (which control the
#               raw -> user coordinate adjustment for lav files) are not
#               available.  It is assumed that endpoint values in the
#               concise file have already been adjusted to user coordinates.
#
#       find_seg
#               Finds the first gap-free segment containing the given
#               position.  Positions falling between gap-free segments
#               within a local alignment are rounded to the lower-indexed
#               segment; gap is set to indicate if this occurred.  Specify
#               a contig header (or prefix) in the second sequence to
#               restrict the segments used, or use NULL to mean "any contig".
#               Searching begins at from_index (normally 0) and goes up
#               from there; this enables repeated calls to find additional
#               matches.  However, it is the RESPONSIBILITY OF THE CALLER
#               to ensure that from_index is a legitimate index for the seg
#               array, i.e., not out-of-bounds.  If no matching segment is
#               found, the return value is -1.
#
#       convert_pos
#               Converts a position from top sequence to bottom, or the
#               inverse, using the segment at the specified index.  It is
#               the RESPONSIBILITY OF THE CALLER to ensure that i is not
#               out-of-bounds.  Setting gap to true indicates that the
#               match fell between this segment and the next, and should
#               be rounded to the end of this segment.  Typically i and
#               gap are the results from a prior call to find_seg().  If
#               i < 0 (no segment found) then the return value is 0.
#
#       has_multi_contig
#               Checks whether the second sequence contained multiple
#               contigs (reverse complement doesn't count).
#
#       get_hdr2 
#               Returns the second-sequence header stored in
#               the contig entry associated with the entry at index i.
#
#       same_local
#               Determines whether two segment entries are in the same
#               local alignment.
#

use 5.6.0;
use strict;
use warnings;
use Class::Struct;


struct Segment => [
    type => '$',	# 's' = contig block, 'a' = local alignment
    			# 'l' = gap-free segment, 'z' = end marker      
    x0 => '$',		# endpoints in first sequence (s/a/l)
    x1 => '$',
    y0 => '$',		# endpoints in second sequence (s/a/l)
    y1 => '$',		
    score => '$',	# score or percent identity (a/l)
    reverse => '$',	# is contig aligned in reverse comp? (s/a/l)
    hdr2 => '$',	# contig header from second sequence (s only)
];


# Constants.

my $LAV = "#:lav";
my $H2PRE = "Aligned to \"";
my $H2SUF = "\":\n";
my $REV1 = "-\"";
my $REV2 = "(reverse complement)";
my $LOCAL = ":";
my $ARROW = "<-->";


# Read the alignment file into the array.

sub get_align {

    (local *ALIGN_FILE, my ($align_file, $inverse)) = @_;
    my $seg_arr;	# Reference to Segment array.

    while (<ALIGN_FILE>) { last unless /^\s*$/; }
    die "$0: \"$align_file\" is empty.\n" if (!defined($_));

    if (/^\s*$LAV/) {
        $seg_arr = read_lav(*ALIGN_FILE, $inverse);
    } elsif (/^\s*$H2PRE/) {
        $seg_arr = read_concise(*ALIGN_FILE, $_);
    } else {
        die "$0: \"$align_file\" is not an alignment file.\n";
    }

    return $seg_arr;



    # Read an alignment file in lav format.

    sub read_lav {

	(local *ALIGN_FILE, my $inverse) = @_;	# Filehandle, Inverse?
        my @buf;				# Input buffer for header.
        my ($x0, $y0, $x1, $y1, $score, $rev);	# Various entries read in from file.
        my $i = -1;				# Segment array index.
	my @seg_arr;			        # Array of Segment structs

        while (<ALIGN_FILE>) {

	    next if (/^\s*$/);
            if (/^\s*d {/) {
		while (<ALIGN_FILE>) { last if /^\s*}/; }
            } elsif (/^\s*s {/) {
                $i++;
		@buf = ();
                die "$0: Unexpected end of lav file.\n" unless defined($_ = <ALIGN_FILE>);
                ($x0, $x1) = get_seqends($_);
                die "$0: Unexpected end of lav file.\n" unless defined($_ = <ALIGN_FILE>);
                ($y0, $y1) = get_seqends($_);
		$rev = (/$REV1/ ? 1 : 0);
                $seg_arr[$i] = new Segment(type=>"s", x0=>$x0, x1=>$x1, y0=>$y0, y1=>$y1, 
					   score=>undef, reverse=>$rev, hdr2=>undef);
            } elsif (/^\s*h {/) {
                die "$0: Unexpected line in lav file:\n$_\n"
                    if ($i < 0 || $seg_arr[$i]->type ne "s");
                while (<ALIGN_FILE>) { last unless /^\s*$/; }
                push (@buf, $_);
                push (@buf, $_ = <ALIGN_FILE>);
                die "$0: Unexpected end of lav file.\n" unless defined($_);
                $seg_arr[$i]->hdr2(get_header_lav($seg_arr[$i]->reverse, $inverse, @buf));

            } elsif (/^\s*a {/) {
                $i++;
                $seg_arr[$i] = new Segment(type=>"x", x0=>undef, x1=>undef, y0=>undef, y1=>undef,
                                           score=>undef, reverse=>undef, hdr2=>undef);
            } elsif (/^\s*s/) {
                die "$0: Unexpected line in lav file:\n$_\n" 
		    if ($i < 0 || $seg_arr[$i]->type ne "x");
                die "$0: Bad line in lav file:\n$_\n" unless ($seg_arr[$i]->score(/^\s*s\s*(\d+)/));
            } elsif (/^\s*b/) {
                die "$0: Unexpected line in lav file:\n$_\n" 
		    if ($i < 0 || $seg_arr[$i]->type ne "x");
                ($x0, $y0) = /^\s*b\s*(\d+)\s+(\d+)/;
                die "$0: Bad line in lav file:\n$_\n" if (!defined($y0));
                $seg_arr[$i]->x0($x0); $seg_arr[$i]->y0($y0);
            } elsif (/^\s*e/) {
                die "$0: Unexpected line in lav file:\n$_\n" 
		    if ($i < 0 || $seg_arr[$i]->type ne "x");
                ($x1, $y1) = /^\s*e\s*(\d+)\s+(\d+)/;
                die "$0: Bad line in lav file:\n$_\n" if (!defined($y1));
                $seg_arr[$i]->x1($x1); $seg_arr[$i]->y1($y1);
                $seg_arr[$i]->type("a");
            } elsif (/^\s*l/) {
                $i++;
                ($x0, $y0, $x1, $y1, $score) =
                     /^\s*l\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/;
                die "$0: Bad line in lav file:\n$_\n" if (!defined($score));
                $seg_arr[$i] = new Segment(type=>"l", x0=>$x0, x1=>$x1, y0=>$y0, y1=>$y1,
				           score=>$score, reverse=>undef, hdr2=>undef);
           }
        }
        $i++;
        $seg_arr[$i] = new Segment(type=>"z", x0=>undef, x1=>undef, y0=>undef, y1=>undef,
                                   score=>undef, reverse=>undef, hdr2=>undef);
        adjust_coords();
        spread_reverse(\@seg_arr);

	return \@seg_arr;


        # Extract sequence endpoints from line in "s {" stanza.

        sub get_seqends {

	    my ($start, $end) =  /^\s*\".*\"\s*(\d+)\s+(\d+)/;
            die "$0: Bad sequence spec in lav file:\n$_\n" unless defined($end);
            return ($start, $end);
        }


        # Extract contig header from line in "h {" stanza; check reverse status.

        sub get_header_lav {

            my ($rev, $inverse, @line) = @_;
	    no warnings;

            die "$0: Bad contig header in lav file:\n$line[0]\n" 
		unless $line[0] =~ /^\s*\".+\"/;
            die "$0: Reverse complement indicators do not agree:\n$line[1]\n\n"
                if (($line[1] !~ /$REV2/ && $rev) || ($line[1] =~ /$REV2/ && !$rev));
            die "$0: Bad contig header in lav file:\n$line[1]\n" 
		unless $line[1] =~ s/^\s*\"(.+)\"\s*/$1/;
	    chomp ($_ = $line[1]);
            s/\($REV2\)//;

            return $_;
        }


        # Convert endpoints from raw blastz coords to user's original sequence.

        sub adjust_coords {

            my ($sx0, $sx1, $sy0, $sy1) = (0, 0, 0, 0);         # Sequence endpoints.
            my $rev = 0;					# Reverse comp?
            my $i;						# Array index.
	    no warnings;

            for ($i = 0; $seg_arr[$i]->type ne "z"; $i++) {
                if ($seg_arr[$i]->type eq "s") {
                    $sx0 = $seg_arr[$i]->x0; $sx1 = $seg_arr[$i]->x1;
                    $sy0 = $seg_arr[$i]->y0; $sy1 = $seg_arr[$i]->y1;
                    $rev = $seg_arr[$i]->reverse;
                } else {
                    $seg_arr[$i]->x0($seg_arr[$i]->x0 + $sx0 - 1);
                    $seg_arr[$i]->x1($seg_arr[$i]->x1 + $sx0 - 1);

                    if ($rev) {
                        $seg_arr[$i]->y0($sy1 + 1 - $seg_arr[$i]->y0);
                        $seg_arr[$i]->y1($sy1 + 1 - $seg_arr[$i]->y1);

                    } else {
                        $seg_arr[$i]->y0($seg_arr[$i]->y0 + $sy0 - 1);
                        $seg_arr[$i]->y1($seg_arr[$i]->y1 + $sy0 - 1);
                    }
                }
            }
        }
    }


    # Read an alignment file in concise format.

    sub read_concise {

        (local *ALIGN_FILE, my $line) = @_;	# Filehandle, first line in concise file.
        my $i = -1;				# Segment Array index.
        my $head;				# Concise file contig header.
	my $rev;				# Reverse status.
        my ($x0, $x1, $y0, $y1, $score);	# Various entries read in from file.
	my @seg_arr = ();	                # Array of Segment structs


        # Read in first header.

        $i++;
        ($head, $rev) = get_header_concise($line);
        $seg_arr[$i] = new Segment(type=>"s", x0=>0, x1=>0, y0=>0, y1=>0,
                                   score=>undef, reverse=>$rev, hdr2=>$head);
        # Deal with the rest of the file.

        while (<ALIGN_FILE>) {

	    next if (/^\s*$/);
            if (/^\s*$H2PRE/) {
                $i++;
                ($head, $rev) = get_header_concise($_);
        	$seg_arr[$i] = new Segment(type=>"s", x0=>0, x1=>0, y0=>0, y1=>0,
                                           score=>undef, reverse=>$rev, hdr2=>$head);
            } elsif (/^\s*$LOCAL/) {
                $i++;
	        $seg_arr[$i] = new Segment(type=>"a", x0=>0, x1=>0, y0=>0, y1=>0,
                                           score=>0, reverse=>undef, hdr2=>undef);
            } elsif (/$ARROW/) {
                $i++;
                ($x0, $x1, $y0, $y1, $score) = /^\s*(\d+)-(\d+) $ARROW (\d+)-(\d+)\s*(\d+)/;
                die "$0: Bad line in concise file:\n$_\n" if (!defined($score));
	        $seg_arr[$i] = new Segment(type=>"l", x0=>$x0, x1=>$x1, y0=>$y0, y1=>$y1,
       	                                   score=>$score, reverse=>undef, hdr2=>undef);
            }
        }
        $i++;
	$seg_arr[$i] = new Segment(type=>"z", x0=>undef, x1=>undef, y0=>undef, y1=>undef,
                                   score=>undef, reverse=>undef, hdr2=>undef);
        spread_reverse(\@seg_arr);
	return \@seg_arr;



        # Extract contig header & reverse status from section line in concise file.

        sub get_header_concise {

            $_ = $_[0];
            my $reverse;

            $reverse = s/ \($REV2\)//;
            die "$0: Bad header format in concise file:\n$_\n" 
		unless (/$H2SUF/ || $reverse);
            s/^\s*$H2PRE//;		# Remove prefix.
            s/$H2SUF\s*$//;             # Remove suffix.

            return ($_, $reverse || 0);
        }
    }


    # Copy reverse values from 's' entries to 'a' and 'l', for convenience.

    sub spread_reverse {

	my $seg_arr = $_[0];		# Reference to Segment array.
        my $rev = 0;			# Reverse comp?
        my $i;				# Array index.

        for ($i = 0; @$seg_arr[$i]->type ne "z"; $i++) {
            if (@$seg_arr[$i]->type eq "s") {
                $rev = @$seg_arr[$i]->reverse;
            } else {
                @$seg_arr[$i]->reverse($rev);
            }
        }
    }
}


# Find the first gap-free segment containing the given position.

sub find_seg {

    my ($pos, $inseq2, $contig, $from_index, $gap, $seg_arr) = @_;
    my $i;					# Array index.
    my $hdr2;					# Sequence header.
    my ($start, $end, $last) = (0, 0, 0);	# Positions in sequence.
    my $key;

    $hdr2 = get_hdr2($from_index, $seg_arr);
    for ($i = $from_index; @$seg_arr[$i]->type ne "z"; $i++) {
        if (@$seg_arr[$i]->type eq "s") {
            $hdr2 = @$seg_arr[$i]->hdr2;
        } elsif (@$seg_arr[$i]->type eq "a") {
            $last = 0;
        } elsif (@$seg_arr[$i]->type eq "l") {
            next if ($contig && $hdr2 !~ /^\Q$contig\E/); 
            $start = ($inseq2 ? @$seg_arr[$i]->y0 : @$seg_arr[$i]->x0);
            $end   = ($inseq2 ? @$seg_arr[$i]->y1 : @$seg_arr[$i]->x1);
            $key = $pos;
            if ($inseq2 && @$seg_arr[$i]->reverse) {
                $start = -$start;
                $end = -$end;
                $key = -$key;
            }
            if ($last && $last < $key && $key < $start) {
                $gap = 1;
                return ($i - 1, $gap);
            } elsif ($start <= $key && $key <= $end) {
                $gap = 0;
                return ($i, $gap);
            } else {
                $last = $end;
            }
        } else {
            die "$0: Internal error: invalid segment type '" . @$seg_arr[$i]->type . "'.\n";
        }
    }
    return (-1, $gap);
}


# Convert a position from top sequence to bottom, or the inverse, using the segment at
# the specified index.

sub convert_pos {

    my ($i, $pos, $inverse, $gap, $seg_arr) = @_;

    if ($i < 0) {
        return;
    } elsif ($gap) {
        return ($inverse ? @$seg_arr[$i]->x1 : @$seg_arr[$i]->y1);
    } elsif ($inverse) {
        return (@$seg_arr[$i]->x0 + (@$seg_arr[$i]->reverse ? -1 : 1) * ($pos - @$seg_arr[$i]->y0));
    } else {
        return (@$seg_arr[$i]->y0 + (@$seg_arr[$i]->reverse ? -1 : 1) * ($pos - @$seg_arr[$i]->x0));
    }
}


# Does the segment overlap?

sub seg_overlaps {

    my ($i, $start, $end, $inseq2, $seg_arr) = @_;
    my ($a0, $a1);

    $a0 = ($inseq2 ? @$seg_arr[$i]->y0 : @$seg_arr[$i]->x0);
    $a1 = ($inseq2 ? @$seg_arr[$i]->y1 : @$seg_arr[$i]->x1);
    ($a0, $a1) = ($a1, $a0) if ($a1 < $a0);
    ($start, $end) = ($end, $start) if ($end < $start);

    return ($start <= $a1 && $end >= $a0);
}


# Checks whether the second sequence contained multiple contigs.

sub has_multi_contig {

    my $seg_arr = $_[0];
    my $hdr2;
    my $i;

    for ($i = 0; @$seg_arr[$i]->type ne "z"; $i++) {
        if (@$seg_arr[$i]->type eq "s") {
            if (!defined($hdr2)) {
                $hdr2 = @$seg_arr[$i]->hdr2;
            } elsif (@$seg_arr[$i]->hdr2 ne $hdr2) {
                return 1;
            }
        }
    }
    return 0;
}


# Returns the second-sequence header stored in the contig entry associated the entry at index i.

sub get_hdr2 {

    my ($i, $seg_arr) = @_;

    for ( ; $i >= 0 && @$seg_arr[$i]->type ne "s"; $i--) { }
    die "$0: Missing sequence header for segment $_[0].\n" if ($i < 0);

    return (@$seg_arr[$i]->hdr2);
}


# Are the two segment entries in the same local alignment?

sub same_local {

    my ($i1, $i2, $seg_arr) = @_;
    my $j;

    ($i1, $i2) = ($i2, $i1) if ($i1 > $i2);
    for ($j = $i1; $j <= $i2; $j++) {
	return 0 if (@$seg_arr[$j]->type ne "l");
    }
    return 1;
}



return 1;
